#include <iostream>
#include <set>

// Bisqwit's 1d Bin Packing algorithm implementation (C++).
// 
// Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)
//
// Read binpacker.hh .

#include "binpacker.hh"

using namespace std;

#define DUMP_ITEMS 1

#define NOWHERE ((unsigned)-1)

namespace
{
    class BinPacker
    {
    public:
        BinPacker
        (
           // Input: List of holes. Value = hole size
           const vector<unsigned> &Holes,
           // Input: List of items. Value = item size
           const vector<unsigned> &Items
        );
        
        const vector<unsigned> GetResult() const;

        void Dump() const;

        bool IsGood() const;
    private:
        struct hole
        {
            int size, free;
            set<unsigned> contents;
        };
        struct item
        {
            unsigned index, location;
            int size;
            bool operator< (const item &b) const { return size > b.size; }
        };
        vector<hole> holes;
        vector<item> items;
        
        void MoveItem(unsigned itemno, unsigned holeno);
        int GetFreeSpace(unsigned holeno) const { return holes[holeno].free; }
        int GetHoleSize(unsigned holeno) const {  return holes[holeno].size; }
        int GetItemSize(unsigned itemno) const {  return items[itemno].size; }
        unsigned GetItemLocation(unsigned itemno) const { return items[itemno].location; }
        
        void Shuffle();
    };

    BinPacker::BinPacker
    (
       const vector<unsigned> &Holes,
       const vector<unsigned> &Items
    ) : holes(Holes.size()),
        items(Items.size())
    {
        for(unsigned a=0; a<Items.size(); ++a)
        {
            items[a].index    = a;
            items[a].size     = Items[a];
            items[a].location = NOWHERE;
        }
        for(unsigned a=0; a<Holes.size(); ++a)
        {
            holes[a].size     = Holes[a];
            holes[a].free     = Holes[a];
        }
        Shuffle();
    }

    const vector<unsigned> BinPacker::GetResult() const
    {
        vector<unsigned> Result(items.size());
        
        for(unsigned a=0; a<items.size(); ++a)
            Result[items[a].index] = items[a].location;

        return Result;
    }

    bool BinPacker::IsGood() const
    {
        for(unsigned a=0; a<holes.size(); ++a)
            if(holes[a].free < 0) return false;
        return true;
    }

    void BinPacker::MoveItem(unsigned itemno, unsigned holeno)
    {
        if(holeno == items[itemno].location) return;
        unsigned oldhole = items[itemno].location;
        if(oldhole != NOWHERE)
        {
            holes[oldhole].free += items[itemno].size;
            holes[oldhole].contents.erase(itemno);
        }
        items[itemno].location = holeno;
        if(holeno != NOWHERE)
        {
            holes[holeno].free -= items[itemno].size;
            holes[holeno].contents.insert(itemno);
        }
    }

    void BinPacker::Shuffle()
    {
        sort(items.begin(), items.end());
        
        for(unsigned a=0; a<items.size(); ++a)
        {
            int leastfree=0; unsigned besthole=0; bool first=true;
            for(unsigned b=0; b<holes.size(); ++b)
                if(holes[b].free >= items[a].size
                && (first || holes[b].free < leastfree)
                  )
                {
                    first = false;
                    besthole = b;
                    leastfree = holes[b].free;
                }
            
            MoveItem(a, besthole);
        }
    }

    void BinPacker::Dump() const
    {
        for(unsigned a=0; a<holes.size(); ++a)
        {
            cerr << "  Hole(size(" << GetHoleSize(a)
                       << ")free(" << GetFreeSpace(a)
                       << ")";
#if DUMP_ITEMS
            cerr << "items(";
            bool first=true,firstout=true;
            unsigned lastitemsize=0,itemcount=0;
            for(set<unsigned>::const_iterator
                i = holes[a].contents.begin();
                i != holes[a].contents.end();
                ++i)
            {
                unsigned size = GetItemSize(*i);
                if(!first && lastitemsize != size)
                {
                    if(firstout)firstout=false;else cerr << ',';
                    cerr << lastitemsize; if(itemcount != 1)cerr << '*' << itemcount;
                }
                if(first || lastitemsize != size)
                {
                    itemcount = 0; lastitemsize = size;
                    first = false;
                }
                ++itemcount;
            }
            if(!first)
            {
                if(!firstout)cerr << ',';
                cerr << lastitemsize;
                if(itemcount != 1)cerr << '*' << itemcount;
            }
            cerr << ')';
#endif
            cerr << ")\n";
        }
    }
}

const vector<unsigned> PackBins
   (const vector<unsigned> &Bins,
    const vector<unsigned> &Items)
{
    BinPacker packer(Bins, Items);
    if(!packer.IsGood()) packer.Dump();
    return packer.GetResult();
}
